/* Huller (online hard margin SVM) implementation.
 *
 * Author: Sebastian Nowozin <nowozin@cs.tu-berlin.de>
 * Date: 22th January 2006.
 *
 * Following the very readable [Border2005].
 *
 * Note: most likely, you want to use a good soft margin SVM implementation
 *     instead.  Only for well separable data this implementation will give
 *     satisfactory results.  See [Schlkopf2002] for details on SVMs in
 *     general.  As alternative implementation, [LibSVM] is known to be a
 *     state-of-the art soft margin SVM implementation.
 *
 * References:
 *
 * [Borders2005]  Bordes, Antoine and Bottou, Leon, 2005, "The Huller: a simple
 *     and efficient online SVM".
 * [Schlkopf2002]  Schlkopf, B. and Smola, A.J., 2002, "Learning with Kernels",
 *     2nd Edition, ISBN 0-262-19475-9
 * [LibSVM] Chih-Chung Chang, Chih-Jen Lin, 2005, "LibSVM - A Library for
 *     Support Vector Machines", http://www.csie.ntu.edu.tw/~cjlin/libsvm/
 *
 *
 * LICENSE
 *
 * Copyright (c) 2006  Sebastian Nowozin <nowozin@cs.tu-berlin.de>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <huller.h>

/* Static function definitions.
 */

/* The huller UPDATE operation, as given in [Border2005].
 */
static void
huller_update (huller* h, unsigned int iidx, int type);

/* Evaluate the kernel function between the two elements.
 *
 * TODO: caching
 *
 * Return the kernel function value.
 */
static float
evaluate_kernel (huller* h, unsigned int idx1, unsigned int idx2);

static void
huller_calculate_X (huller* h);

/* Code
 */

huller *
huller_create (unsigned int samples_count, unsigned char* samples_labels,
	float (* kernel)(unsigned int, unsigned int))
{
	huller*	h;
	unsigned int	n;

	h = calloc (1, sizeof (huller));
	h->samples_count = samples_count;
	h->samples_labels = samples_labels;
	h->kernel = kernel;

	/* Create index arrays, alphas.
	 */
	h->non_sv_index = calloc (samples_count, sizeof (unsigned int));
	h->sv_index = calloc (samples_count, sizeof (unsigned int));
	h->alpha = calloc (samples_count, sizeof (float));

	/* Put all indices into the non_sv index array.
	 */
	for (n = 0 ; n < samples_count ; ++n)
		h->non_sv_index[n] = n;

	h->non_sv_index_count = samples_count;

	return (h);
}


void
huller_destroy (huller* h)
{
	free (h->alpha);
	free (h->sv_index);
	free (h->non_sv_index);

	free (h);
}


void
huller_dump (huller* h)
{
	int n;

	printf ("Support vectors (%u pos, %u neg):\n",
		h->sv_index_count_pos, h->sv_index_count_neg);
	for (n = 0 ; n < h->sv_index_count ; ++n) {
		printf ("\t%u\t%c\t%f\n", h->sv_index[n],
			"-+"[h->samples_labels[h->sv_index[n]]],
			h->alpha[h->sv_index[n]]);
	}

	printf ("Internal state:\n");
	printf ("    XPXP = %f, XNXP = %f, XNXN = %f\n",
		h->M_XPXP, h->M_XNXP, h->M_XNXN);
	printf ("\n");
}


float
huller_evaluate (huller* h)
{
	float bias = (h->M_XNXN - h->M_XPXP)/2.0;
	float pos_sum = 0.0;
	float neg_sum = 0.0;
	float val;	// Temporary
	unsigned int n;

	for (n = 0 ; n < h->sv_index_count ; ++n) {
		int idx = h->sv_index[n];

		val = h->alpha[idx] * evaluate_kernel (h, idx, HULLER_TEST_SAMPLE);
		if (h->samples_labels[idx] == 0) {
			// negative
			neg_sum += val;
		} else {
			pos_sum += val;
		}
	}

	return (pos_sum - neg_sum + bias);
}

/* Initialize the first few support vectors randomly by averaging a few
 * points.
 *
 * Return 0 on success, -1 in case not enough samples are there to use the
 * specified number of boot samples.
 */
int
huller_initialize (huller* h, unsigned int boot_sample_count)
{
	unsigned int samples_done;
	unsigned int sample_start;
	unsigned int sample_iter;
	unsigned int pn_iter;
	float alpha = 1.0 / ((float) boot_sample_count);

	for (pn_iter = 0 ; pn_iter <= 1 ; ++pn_iter) {
		samples_done = 0;
		do {
			sample_iter = sample_start =
				random () % h->non_sv_index_count;

			do {
				/* Sample matches the required label.
				 */
				int idx = h->non_sv_index[sample_iter];

				if (h->samples_labels[idx] == pn_iter) {
					h->sv_index[h->sv_index_count] = idx;
					h->sv_index_count += 1;
					UPDATE_POSNEG_COUNT (pn_iter, 1);
					h->alpha[idx] = alpha;

					h->non_sv_index_count -= 1;
					h->non_sv_index[sample_iter] =
						h->non_sv_index[h->non_sv_index_count];

					samples_done += 1;

					goto next_sample;
				}

				sample_iter += 1;
				sample_iter %= h->samples_count;
			} while (sample_iter != sample_start);

			if (sample_start == sample_iter)
				return (-1);

next_sample:
			;
		} while (samples_done < boot_sample_count);
	}

	huller_calculate_X (h);

	return (0);
}


static void
huller_calculate_X (huller* h)
{
	unsigned int n, n2;

	/* Calculate the initial M_XPXP, M_XNXN and M_XNXP
	 */
	h->M_XPXP = h->M_XNXN = h->M_XNXP = 0;

	for (n = 0 ; n < h->sv_index_count ; ++n) {
		unsigned int idx = h->sv_index[n];
		int label = h->samples_labels[idx];

		for (n2 = 0 ; n2 < h->sv_index_count ; ++n2) {
			unsigned int idx2 = h->sv_index[n2];

			if (label == 1 && h->samples_labels[idx2] == 0) {
				h->M_XNXP += h->alpha[idx] * h->alpha[idx2] *
					evaluate_kernel (h, idx, idx2);
			} else if (label == h->samples_labels[idx2]) {
				float val = h->alpha[idx] * h->alpha[idx2] *
					evaluate_kernel (h, idx, idx2);

				if (label == 1)
					h->M_XPXP += val;
				else
					h->M_XNXN += val;
			}
		}
	}
}


/* Iterates a whole epoch.
 */
void
huller_iterate_epoch (huller* h)
{
	/* Produce a copy of the current non SV index array.
	 */
	unsigned int*	non_sv_index_fix;
	unsigned int	non_sv_index_fix_count;

	non_sv_index_fix = calloc (h->non_sv_index_count,
		sizeof (unsigned int));
	memcpy (non_sv_index_fix, h->non_sv_index,
		h->non_sv_index_count * sizeof (unsigned int));
	non_sv_index_fix_count = h->non_sv_index_count;

	/* Now, for all non SV's, try to insert them.  Then randomly try to
	 * remove any current SV.
	 */
	do {
		unsigned int idx, iidx;
		unsigned int iidx_start;

		iidx_start = iidx = random () % non_sv_index_fix_count;
		idx = non_sv_index_fix[iidx];

		/* Search for the original idx in the original array.
		 * First try the old iidx, which will work well for the first
		 * half of the non-SV's, but later we will linearly search
		 * from there.
		 * This could be quite an expensive operation, maybe we should
		 * change the data structure.
		 *
		 * Note: the idx index is guaranteed to appear in the original
		 *       non_sv_index, as the only way it could leave from
		 *       there is by pushing it as support vector.  But only
		 *       elements we choose can be pushed.
		 */
		while (h->non_sv_index[iidx] != idx) {
			iidx += 1;
			iidx %= h->non_sv_index_count;
			if (iidx == iidx_start) {
				printf ("FATAL: non-SV slipped from non_sv_index\n");
				printf ("       iidx = %u, idx = %u\n", iidx, idx);

				printf ("Dumping h->non_sv_index:\n");
				for (iidx = 0 ; iidx < h->non_sv_index_count ; ++iidx)
					printf ("    %u\t%u\n", iidx,
						h->non_sv_index[iidx]);

				exit (EXIT_FAILURE);
			}
		}

		/* Do an update with a random non-sv, followed by an update
		 * with a random support vector.
		 */
		huller_update (h, iidx, HULLER_TYPE_NONSV);

		if (h->sv_index_count_pos > 1 || h->sv_index_count_neg > 1) {
			do {
				unsigned int sv_iidx = random () % h->sv_index_count;
				unsigned int count =
					h->samples_labels[h->sv_index[sv_iidx]] == 1 ?
					h->sv_index_count_pos : h->sv_index_count_neg;

				/* Wrong subset, try again.  As this subset
				 * has exactly one element, chances are > 2/3
				 * to hit the correct subset next time.
				 */
				if (count <= 1)
					continue;

				huller_update (h, sv_iidx, HULLER_TYPE_SV);
				break;
			} while (1);
		}

		/* Squash processed non-SV with the last element, making our
		 * copy of the old non-SV's shorter every iteration.
		 */
		non_sv_index_fix_count -= 1;
		non_sv_index_fix[iidx_start] =
			non_sv_index_fix[non_sv_index_fix_count];
	} while (non_sv_index_fix_count > 0);

	free (non_sv_index_fix);
}


/* One iteration step, consisting of two update calls.
 */
void
huller_iterate (huller* h)
{
	unsigned int sv_iidx;
	unsigned int count;

	if (h->non_sv_index_count > 0)
		huller_update (h, random () % h->non_sv_index_count, HULLER_TYPE_NONSV);

#ifdef	DEBUG
	printf ("\n");
	huller_dump (h);
	printf ("\n");
#endif

	/* Never make the support vector set empty for positive or negative
	 * samples.
	 */

	if (h->sv_index_count_pos <= 1 && h->sv_index_count_neg <= 1)
		return;

	do {
		sv_iidx = random() % h->sv_index_count;
		count = h->samples_labels[h->sv_index[sv_iidx]] == 1 ?
			h->sv_index_count_pos : h->sv_index_count_neg;

		if (count > 1) {
			huller_update (h, sv_iidx, HULLER_TYPE_SV);

			return;
		}
	} while (1);
}


static void
huller_update (huller* h, unsigned int iidx, int type)
{
	float XPxk = 0.0;
	float XNxk = 0.0;
	float xkxk;
	float lambda;	// lambda, first unconstrained, then constrained
	float alpha_exp;
	float lambda_scale;
	unsigned int n;
	unsigned int idx;
	int label;

	if (type == HULLER_TYPE_SV) {
		idx = h->sv_index[iidx];
	} else {
		idx = h->non_sv_index[iidx];
	}
#ifdef	DEBUG
	printf ("UPDATE: trying to %s sample %u\n",
		type == HULLER_TYPE_SV ? "prune" : "merge", idx);
#endif

	/* K(x_k, x_k)
	 */
	xkxk = evaluate_kernel (h, idx, idx);

	/* X_P x_k and
	 * X_N x_k at the same time
	 */
	for (n = 0 ; n < h->sv_index_count ; ++n) {
		unsigned int s_idx = h->sv_index[n];

		if (h->samples_labels[s_idx] == 0) {
			// Negative sample
			XNxk += h->alpha[s_idx] * evaluate_kernel (h, s_idx, idx);
		} else {
			// Positive sample
			XPxk += h->alpha[s_idx] * evaluate_kernel (h, s_idx, idx);
		}
	}

	/*
	printf ("h->M_XNXN - xkxk - 2.0*XNxk = "
		"%f - %f - 2.0*%f = %f\n",
		h->M_XNXN, xkxk, XNxk, h->M_XNXN - xkxk - 2.0*XNxk);
	printf ("h->M_XPXP + xkxk - 2.0*XPxk = "
		"%f + %f - 2.0*%f = %f\n",
		h->M_XPXP, xkxk, XPxk, h->M_XPXP + xkxk - 2.0*XPxk);
	*/

	if (h->samples_labels[idx] == 0) {
#ifdef	DEBUG
		// negative sample
		if (fabsf (h->M_XNXN + xkxk - 2.0*XNxk) < HULLER_ZERO_EPSILON) {
			printf ("negative DEGENERATE:\n");
			huller_dump (h);
			printf ("current index: %d\n", idx);
			printf ("current type: %s\n",
				(type == HULLER_TYPE_SV) ? "TYPE_SV" : "TYPE_NON_SV");
		}

		printf ("   Unconstrainted lambda calculation (negative sample)\n");
		printf ("      (h->M_XNXN - h->M_XNXP - XNxk + XPxk) / (h->M_XNXN - xkxk - 2.0*XNxk)\n");
		printf ("      = (%f - %f - %f + %f) / (%f + %f - 2.0*%f)\n",
			h->M_XNXN, h->M_XNXP, XNxk, XPxk, h->M_XNXN, xkxk, XNxk);
#endif

		lambda = (h->M_XNXN - h->M_XNXP - XNxk + XPxk) /
			(h->M_XNXN + xkxk - 2.0*XNxk);

#ifdef	DEBUG
		printf ("      = %f\n", lambda);
#endif
	} else {
#ifdef	DEBUG
		// positive sample
		if (fabsf (h->M_XPXP + xkxk - 2.0*XPxk) < HULLER_ZERO_EPSILON) {
			printf ("positive DEGENERATE:\n");
			huller_dump (h);
			printf ("current index: %d\n", idx);
			printf ("current type: %s\n",
				(type == HULLER_TYPE_SV) ? "TYPE_SV" : "TYPE_NON_SV");
		}

		printf ("   Unconstrainted lambda calculation (positive sample)\n");
		printf ("      (h->M_XPXP - h->M_XNXP - XPxk + XNxk) / (h->M_XPXP + xkxk - 2.0*XPxk)\n");
		printf ("      = (%f - %f - %f + %f) / (%f + %f - 2.0*%f)\n",
			h->M_XPXP, h->M_XNXP, XPxk, XNxk, h->M_XPXP, xkxk, XPxk);
#endif

		lambda = (h->M_XPXP - h->M_XNXP - XPxk + XNxk) /
			(h->M_XPXP + xkxk - 2.0*XPxk);

#ifdef	DEBUG
		printf ("      = %f\n", lambda);
#endif
	}

	/* Calculate the optimal lambda, using a fix by Haffner to allow the
	 * alpha_k to become zero in one update step instead of only slowly
	 * decreasing.
	 */
	//printf ("h->alpha[%d] = %f\n", idx, h->alpha[idx]);
	if (fabsf (h->alpha[idx] - 1) < HULLER_ZERO_EPSILON) 
		alpha_exp = -9999.0;
	else
		alpha_exp = -h->alpha[idx] / (1.0 - h->alpha[idx]);

#ifdef	DEBUG
	printf ("   Lambda calculation\n");
	printf ("      alpha_exp = %f\n", alpha_exp);
	printf ("      lambda_u = %f\n", lambda);
#endif
	lambda = alpha_exp > lambda ? alpha_exp : lambda;
	lambda = 1.0 < lambda ? 1.0 : lambda;
#ifdef	DEBUG
	printf ("      lambda = %f\n", lambda);
#endif

	/* Nothing to do if lambda is zero.
	 */
	if (fabsf (lambda) < HULLER_ZERO_EPSILON)
		return;

	/* Adjust all the alpha values for all support vectors.
	 * After the first adjustment the sum of the alpha values is 1.0 - lambda,
	 * hence we have lambda free to add to any element of choice.
	 */
	lambda_scale = 1.0 - lambda;
	label = h->samples_labels[idx];

#ifdef	DEBUG
	printf ("   Reducing SV weights\n");
#endif
	for (n = 0 ; n < h->sv_index_count ; ++n) {
		int sv_idx = h->sv_index[n];

		if (h->samples_labels[sv_idx] != label)
			continue;

#ifdef	DEBUG
		printf ("      h->alpha[%d] = %f * %f = %f\n", sv_idx, h->alpha[sv_idx],
			lambda_scale, lambda_scale * h->alpha[sv_idx]);
#endif
		h->alpha[sv_idx] *= lambda_scale;

		/* Here its also necessary to check for degenerating alpha's.
		 * [Bordes2005] does not discuss this.
		 *
		 * The support vector we process in this UPDATE operation we
		 * can safely skip from removing, as we will add lambda to it
		 * just below this for loop.
		 */
		if (sv_idx != idx && fabsf (h->alpha[sv_idx]) < HULLER_ZERO_EPSILON) {
			// Remove index from support vector index array
			h->sv_index_count -= 1;
			h->sv_index[n] = h->sv_index[h->sv_index_count];
			UPDATE_POSNEG_COUNT (h->samples_labels[sv_idx], -1);

			// Add index to non support vector array
			h->non_sv_index[h->non_sv_index_count] = sv_idx;
			h->non_sv_index_count += 1;

			n -= 1;	// Do the moved element again.
		}
	}

	h->alpha[idx] += lambda;
#ifdef	DEBUG
	printf ("   NEW: h->alpha[%d] = %f (type: %s)\n", idx, h->alpha[idx],
		type == HULLER_TYPE_SV ? "TYPE_SV" : "TYPE_NON_SV");
#endif

	if (type == HULLER_TYPE_SV) {
		// A support vector can become a non support vector by having its
		// alpha value changed to zero.
		if (fabsf (h->alpha[idx]) < HULLER_ZERO_EPSILON) {
			// Remove index from support vector index array
			h->sv_index_count -= 1;
			h->sv_index[iidx] = h->sv_index[h->sv_index_count];
			UPDATE_POSNEG_COUNT (h->samples_labels[idx], -1);

			// Add index to non support vector array
			h->non_sv_index[h->non_sv_index_count] = idx;
			h->non_sv_index_count += 1;
		}
	} else {
		// A non support vector can become a support vector by having an alpha
		// larger than zero
		if (fabsf (h->alpha[idx]) >= HULLER_ZERO_EPSILON) {
			// Remove index from non support vector index array
			h->non_sv_index_count -= 1;
			h->non_sv_index[iidx] = h->non_sv_index[h->non_sv_index_count];

			// Add index to support vector array
			h->sv_index[h->sv_index_count] = idx;
			h->sv_index_count += 1;
			UPDATE_POSNEG_COUNT (h->samples_labels[idx], 1);
		}
	}

	/* Now update the managed variables
	 */
	if (h->samples_labels[idx] == 0) {
		// negative sample
		h->M_XNXP = lambda_scale * h->M_XNXP + lambda * XPxk;
		h->M_XNXN = (lambda_scale*lambda_scale) * h->M_XNXN +
			2.0 * lambda * lambda_scale * XNxk +
			lambda * lambda * xkxk;
	} else {
		// positive sample
		h->M_XPXP = (lambda_scale*lambda_scale) * h->M_XPXP +
			2.0 * lambda * lambda_scale * XPxk +
			lambda * lambda * xkxk;
		h->M_XNXP = lambda_scale * h->M_XNXP + lambda * XNxk;
	}
}


static float
evaluate_kernel (huller* h, unsigned int idx1, unsigned int idx2)
{
	// TODO: kernel value caching
	return (h->kernel (idx1, idx2));
}


