/* Huller (online hard margin SVM) implementation.
 *
 * Author: Sebastian Nowozin <nowozin@cs.tu-berlin.de>
 * Date: 22th January 2006.
 *
 * For information how to use the interface defined in this header file,
 * please see the included example.c file.
 *
 * http://cs.tu-berlin.de/~nowozin/huller/
 */

#ifndef	HULLER_H
#define	HULLER_H

#define	HULLER_VERSION	"0.1"

/* The eps to decide when an element leaves the support vector set.
 */
#define	HULLER_ZERO_EPSILON	1e-6


/* The basic huller object every other function operates on.
 * Created using huller_create and free'd using huller_destroy.
 */
typedef struct {
	/* The number of training samples.
	 * non_sv_index_count + sv_index_count == samples_count.
	 */
	unsigned int	samples_count;
	/* The labels of the training samples.  Must be 0 or 1.
	 * 0 is in the negative set N, 1 is in the positive set P.
	 */
	unsigned char*	samples_labels;

	/* The kernel function, taking two indexes.  You will have to supply a
	 * kernel function of your choice here.
	 * Note: if the _second_ index is set to HULLER_TEST_SAMPLE the kernel
	 * function should return k(i1, testsample).
	 */
	float (* kernel)(unsigned int, unsigned int);
#define	HULLER_TEST_SAMPLE	0xffffffff

#define	HULLER_TYPE_NONSV	0
	unsigned int	non_sv_index_count;
	unsigned int*	non_sv_index;

#define	HULLER_TYPE_SV		1
	unsigned int	sv_index_count;
	unsigned int*	sv_index;
	unsigned int	sv_index_count_pos;
	unsigned int	sv_index_count_neg;
#define	UPDATE_POSNEG_COUNT(label,adj) { \
	if ((label) == 0) \
		h->sv_index_count_neg += (adj); \
	else \
		h->sv_index_count_pos += (adj); \
}

	/* The full alpha vector for all samples.
	 */
	float*	alpha;

	/* The managed variables we need very often.
	 */
	float	M_XPXP;	// X_P X_P
	float	M_XNXP;	// X_N X_P
	float	M_XNXN;	// X_N X_N
} huller;


/* Create a new huller object.
 *
 * The huller will be trained on 'samples_count' samples, with the associated
 * labels 'samples_labels' (read-only), and the 'kernel' function as kernel.
 *
 * Return a new huller object.
 */
huller *
huller_create (unsigned int samples_count, unsigned char* samples_labels,
	float (* kernel)(unsigned int, unsigned int));

/* Destroy a huller object.
 *
 * Free all resources associated with the huller object.
 */
void
huller_destroy (huller* h);

/* Debugging function: dumping the huller state.
 */
void
huller_dump (huller* h);

/* Initialize the first few support vectors randomly by averaging a few
 * points.
 *
 * Return 0 on success, -1 in case not enough samples are there to use the
 * specified number of boot samples.
 */
int
huller_initialize (huller* h, unsigned int boot_sample_count);

/* Iterates a whole epoch.
 *
 * This tries to push all non support vectors into the support vector set
 * once.
 */
void
huller_iterate_epoch (huller* h);

/* One huller iteration step, consisting of two update calls.
 *
 * Updates a single random non-SV and a single random SV.
 */
void
huller_iterate (huller* h);

/* Evaluate the function defined by the support vector and bias for a single
 * test sample.
 *
 * This is done by calling the kernel function with a special index
 * (HULLER_TEST_SAMPLE).  So you have to modify the kernel function
 * accordingly to check for the special index.
 *
 * Return the function value for the given sample.
 */
float
huller_evaluate (huller* h);

#endif


